#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "base.h"
#include "log.h"
#include "buffer.h"

#include "plugin.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

typedef struct {
	pcre_keyvalue_buffer *rewrite;
} plugin_config;

typedef struct {
	PLUGIN_DATA;
	buffer *match_buf;
	
	plugin_config **config_storage;
	
	plugin_config conf; 
} plugin_data;

INIT_FUNC(mod_rewrite_init) {
	plugin_data *p;
	
	p = calloc(1, sizeof(*p));
	
	p->match_buf = buffer_init();
	
	return p;
}

FREE_FUNC(mod_rewrite_free) {
	plugin_data *p = p_d;
	
	UNUSED(srv);

	if (!p) return HANDLER_GO_ON;
	
	buffer_free(p->match_buf);
	if (p->config_storage) {
		size_t i;
		for (i = 0; i < srv->config_context->used; i++) {
			plugin_config *s = p->config_storage[i];
			pcre_keyvalue_buffer_free(s->rewrite);
			
			free(s);
		}
		free(p->config_storage);
	}
	
	free(p);
	
	return HANDLER_GO_ON;
}

SETDEFAULTS_FUNC(mod_rewrite_set_defaults) {
	plugin_data *p = p_d;
	data_unset *du;
	data_config *dc;
	size_t i = 0;
	int ret = HANDLER_GO_ON;
	
	config_values_t cv[] = { 
		{ "url.rewrite",                NULL, T_CONFIG_LOCAL, T_CONFIG_SCOPE_CONNECTION }, /* 0 */
		{ NULL,                         NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET }
	};
	
	if (!p) return HANDLER_ERROR;
	
	/* 0 */
	p->config_storage = malloc(srv->config_context->used * sizeof(specific_config *));
	
	for (i = 0; i < srv->config_context->used; i++) {
		plugin_config *s;
		size_t j;
		data_array *da = (data_array *)du;
		
		s = malloc(sizeof(plugin_config));
		s->rewrite   = pcre_keyvalue_buffer_init();
		
		cv[0].destination = s->rewrite;
		
		p->config_storage[i] = s;
		srv->config = ((data_config *)srv->config_context->data[i])->value;
	
		if (0 != (ret = config_insert_values(srv, cv))) {
			ret = HANDLER_ERROR;
			
			break;
		}
		
		ret = HANDLER_GO_ON;
		
		if (NULL == (du = array_get_element(srv->config, "url.rewrite"))) {
			/* no url.rewrite defined */
			continue;
		}
		
		if (du->type != TYPE_ARRAY) {
			log_error_write(srv, __FILE__, __LINE__, "sss", 
					"unexpected type for key: ", "url.rewrite", "array of strings");
			
			return HANDLER_ERROR;
		}
		
		da = (data_array *)du;
				
		for (j = 0; j < da->value->used; j++) {
			if (da->value->data[j]->type != TYPE_STRING) {
				log_error_write(srv, __FILE__, __LINE__, "sssbs", 
						"unexpected type for key: ", 
						"url.rewrite", 
						"[", da->value->data[j]->key, "](string)");
				
				return HANDLER_ERROR;
			}
				
			if (0 != pcre_keyvalue_buffer_append(s->rewrite, 
							     ((data_string *)(da->value->data[j]))->key->ptr,
							     ((data_string *)(da->value->data[j]))->value->ptr)) {
					
				log_error_write(srv, __FILE__, __LINE__, "sb", 
						"pcre-compile failed for", da->value->data[j]->key);
			}
		}
	}
	
	if (NULL != (dc = (data_config *)array_get_element(srv->config_context, "global"))) {
		srv->config = dc->value;
	}
	
	return ret;
}

static int mod_rewrite_patch_connection(server *srv, connection *con, plugin_data *p, const char *stage) {
	size_t i, j;
	
	/* skip the first, the global context */
	for (i = 1; i < srv->config_context->used; i++) {
		data_config *dc = (data_config *)srv->config_context->data[i];
		plugin_config *s = p->config_storage[i];
		
		/* not our stage */
		if (0 != strcmp(dc->comp_key->ptr, stage)) continue;
		
		/* condition didn't match */
		if (!config_check_cond(srv, con, dc)) continue;
		
		/* merge config */
		for (j = 0; j < dc->value->used; j++) {
			data_unset *du = dc->value->data[j];
			
			if (0 == strcmp(du->key->ptr, "url.rewrite")) {
				p->conf.rewrite = s->rewrite;
			}
		}
	}
	
	return 0;
}

static int mod_rewrite_setup_connection(server *srv, connection *con, plugin_data *p) {
	plugin_config *s = p->config_storage[0];
	
	UNUSED(srv);
	UNUSED(con);
		
	p->conf.rewrite = s->rewrite;
	
	return 0;
}




URIHANDLER_FUNC(mod_rewrite_uri_handler) {
#ifdef HAVE_PCRE_H
	plugin_data *p = p_d;
	size_t i;

	/* 
	 * REWRITE URL
	 * 
	 * e.g. rewrite /base/ to /index.php?section=base
	 * 
	 */
	
	mod_rewrite_setup_connection(srv, con, p);
	mod_rewrite_patch_connection(srv, con, p, "HTTPhost");
	mod_rewrite_patch_connection(srv, con, p, "HTTPurl");
	
	buffer_copy_string_buffer(p->match_buf, con->request.uri);
	
	for (i = 0; i < p->conf.rewrite->used; i++) {
		pcre *match;
		const char *pattern;
		size_t pattern_len;
		int n;
# define N 10
		int ovec[N * 3];
		
		match = p->conf.rewrite->kv[i]->key;
		pattern = p->conf.rewrite->kv[i]->value;
		pattern_len = strlen(pattern);
		
		if ((n = pcre_exec(match, NULL, p->match_buf->ptr, p->match_buf->used - 1, 0, 0, ovec, 3 * N)) < 0) {
			if (n != PCRE_ERROR_NOMATCH) {
				log_error_write(srv, __FILE__, __LINE__, "sd",
						"execution error while matching: ", n);
				return HANDLER_ERROR;
			}
		} else {
			const char **list;
			size_t start, end;
			size_t k;
			
			/* it matched */
			pcre_get_substring_list(p->match_buf->ptr, ovec, n, &list);
			
			/* search for $[0-9] */
			
			buffer_reset(con->request.uri);
			
			start = 0; end = pattern_len;
			for (k = 0; k < pattern_len; k++) {
				if (pattern[k] == '$' &&
				    isdigit((unsigned char)pattern[k + 1])) {
					/* got one */
					
					size_t num = pattern[k + 1] - '0';
					
					end = k;
					
					buffer_append_string_len(con->request.uri, pattern + start, end - start);
					
					/* n is always larger than 0 */
					if (num < (size_t)n) {
						buffer_append_string(con->request.uri, list[num]);
					}
					
					k++;
					start = k + 1;
				} 
			}
			
			buffer_append_string_len(con->request.uri, pattern + start, pattern_len - start);
			
			pcre_free(list);
			
			return HANDLER_COMEBACK;
		}
	}
#undef N
		
#else
	UNUSED(srv);
	UNUSED(con);
	UNUSED(p_d);
#endif

	return HANDLER_GO_ON;
}

int mod_rewrite_plugin_init(plugin *p) {
	p->name        = buffer_init_string("rewrite");
	
	p->init        = mod_rewrite_init;
	p->handle_uri_raw = mod_rewrite_uri_handler;
	p->set_defaults = mod_rewrite_set_defaults;
	p->cleanup     = mod_rewrite_free;
	
	p->data        = NULL;
	
	return 0;
}
